"""CheckType classes utility module.

This module contains:

* CheckType   : superclass to be inherited from (object oriented API)
* checktype   : factory building a CheckType subclass (functional API)
* checktyped  : class decorator adding checktype descriptors
* Validator   : descriptor class performing typechecking

A CheckType is a type providing a custom behaviour when passed in the
second argument of isinstance(): it will return True if it first
argument satisfies a predicate and False otherwise. This predicate is a
class callable named predicate.

If a CheckType inherits from other classes, it will check first if the
passed object is an instance of all of them.

A CheckType can also be also be instantiated and used as a data descriptor
using the descriptor() alternative constructorinside a class body: it returns
an instance of the Validator class, wich is a descriptor, and an exception will
be raised when trying to assign a value that doesn't pass the isinstance()
check. See CheckType.validate documentation.

A value can be converted into a CheckType if it passes the isinstance() check.
"""

from abc import ABCMeta, abstractmethod
from reprlib import repr
from keyword import iskeyword


class _MetaCheckType(ABCMeta):
    """Metaclass for CheckTypes.

    It inhertits from ABCMeta to avoid metaclass conflicts.
    """

    def __new__(mcls, name, bases, ns):
        cls = type.__new__(mcls, name, bases, ns)  # Inhibit ABCMeta.__new__
        cls._checktype_registry = list()
        doc = cls.__doc__
        clsname = cls.__name__
        expected = f"{clsname!r} ({doc})" if doc else f"{clsname!r}"
        cls._errmsg1 = f"expected {expected} but got ""{}"
        cls.__repr__ = cls.__class__._repr
        return cls

    def __instancecheck__(cls, instance):
        # Override for isinstance() when the second argument is a CheckType
        try:
            cls._CheckType__validate(instance)
        except:  # Any kind of exception is interpreted as a failure
            if not cls._check_registries(instance):
                return False
        return True

    def _check_registries(cls, instance):
        # Recursive helper method checking the registry of a CheckType
        # and of all of its subclasses.
        if cls is CheckType:
            return True
        if instance in cls._checktype_registry:
            return True
        for subcls in cls.__subclasses__():
            if subcls._check_registries(instance):
                # Cache the instance in the registry of the original class if
                # it was found in one of its subclasses.
                cls.register(instance)
                return True
        return False

    def _dump_registry(cls, file=None):
        # Debug helper
        print("Class: %s.%s" % (cls.__module__, cls.__qualname__), file=file)
        for attr in cls.__dict__:
            if attr.startswith('_checktype_'):
                print(f'{attr}: {getattr(cls, attr)}', file=file)

    @staticmethod
    def _repr(self):
        # __repr__ prototype for MetaCheckType class instances.
        cls = self.__class__
        clsname = cls.__name__
        return f"{clsname}({cls.__base__.__repr__(self)})"

    # Inhibit ABCMeta issubclass() override
    __subclasscheck__ = type.__subclasscheck__


class CheckType(metaclass=_MetaCheckType):
    """Special kind of type providing a type checking abstraction."""

    def __init__(self, *args, **kwargs):
        """Raise an exception if the value doesn't satisfy
        self.validate().
        """
        if args or kwargs:
            try:
                self.validate(self)
            except Exception as e:
                raise type(e)(f"could not convert value") from e

    @classmethod
    def descriptor(cls, owner=None, name=None):
        """Instantiate a CheckType as a descriptor.

        Args:
            owner: the class that is targeted to own the descriptor.
            name: the name to which the descriptor is aimed to be
                assigned in the namespace of the class.

        Returns:
            An instance of the Validator class.
        """
        return Validator(type=cls, owner=owner, name=name)

    @staticmethod
    def predicate(val):
        """Callable testing a value.

        This method is meant to be overriden by the subclasser.
        It can be any callable assigned to a class attribute, a
        staticmethod or a classmethod.
        This method is not meant to be used after the class creation,
        but other methods will leverage it for validation purpose.

        Args:
            val: any value to test.

        Returns:
            A boolean (or value aimed to be evaluated as a boolean) that
            equals True if the value passed the test, False otherwise.

        Notes:
            In the CheckType superclass, the default implementation is
            to always return True.
        """
        return True

    @classmethod
    def validate(cls, val):
        """Data validation classmethod.

        It will raise an appropriate exception according to the value
        passed as argument.

        Args:
            val: any value to test.

        Raises:
            TypeError: the value is not an instance of all cls bases.
            ValueError: the value doesn't satisfy predicate().
        """
        try:
            cls.extra_validation(val)
            if not all(isinstance(val, base) for base in cls.__bases__):
                raise TypeError(cls.get_errmsg(type(val)))
            if not cls.predicate(val):
                raise ValueError(cls.get_errmsg(val))
        except:
            if not cls._check_registries(val):
                raise

    __validate = validate  # Prevent from override in subclasses

    @classmethod
    def extra_validation(cls, val):
        """Extra validation class method called by validate.

        This method is meant to be override in subclasses in order to
        perform extra validation or raise different kind of exceptions.
        If a TypeError or a ValueError is raised, the error message will
        be handled automatically, unless one is already given.

        Args:
            val: any value to test.

        Raises:
            up to the subclasser...
        """
        pass

    @classmethod
    def register(cls, instance):
        """Register a virtual instance of a CheckType.

        This classmethod can be used to force some value to pass the
        isinstance() and validate() tests.
        Be aware that a value passed in the registry of a subclass will
        also pass the tests of all of its superclasses.

        Args:
            instance: any value.
        """
        if instance not in cls._checktype_registry:
            cls._checktype_registry.append(instance)

    @classmethod
    def get_errmsg(cls, but_got):
        """Get an preformatted error message from a value.

        Args:
            but_got: non expected received value (if it's a class its
                name will be used).

        Returns:
            An error message as a string that looks like:
            "expected '<SomeClass>' but got <some_val>"
        """
        if isinstance(but_got, type):
            but_got = but_got.__name__
        return cls._errmsg1.format(repr(but_got))


def checktype(name, bases=None, predicate=None, doc=None, **kwargs):
    """Factory function returning a CheckType subclass.

    Args:
        name: the name of the class to be returned.
        bases: the class or tuple of classes to be inherited from.
        predicate: a callable, staticmethod or classmethod accepting one
            and only argument and returning a boolean.
        doc: the __doc__ of the class to be returned.
        kwargs: attributes or methods to be injected in the namespace of
            the class.

    Returns:
        A subclass of CheckType.
    """
    if bases is None:
        bases = (CheckType,)
    elif isinstance(bases, tuple):
        bases = (*bases, CheckType)
    elif isinstance(bases, type):
        bases = (bases, CheckType)
    namespace = {'__doc__': doc, **kwargs}
    if predicate:
        namespace['predicate'] = predicate
    return type(name, bases, namespace)


def checktyped(cls):
    """Decorator for injecting validators in a class.

    It uses the special attribute __annotations__  of the class if
    found. Every annotation with a non type value will be ignored.

    Args:
        cls: the class to decorate.

    Returns:
        The same class with validators.
    """
    annotations = cls.__dict__.get('__annotations__')
    if annotations:
        for name, typ in annotations.items():
            if isinstance(typ, type):
                setattr(cls, name, Validator(type=typ, owner=cls, name=name))
    return cls


class Validator:
    """Descriptor class performing type type checking."""

    def __init__(self, type=object, owner=None, name=None):
        self.type = type
        if owner or name:
            self._check_owner_and_name(owner, name)
            self.__set_name__(owner, name)

    def _check_owner_and_name(self, owner, name):
        if not isinstance(owner, type):
            raise TypeError("'owner' must be a class")
        if (not isinstance(name, str) or
            not name.isidentifier() or iskeyword(name)):
            raise TypeError("'name' must be a valid identifier")

    def __set_name__(self, owner, name):
        self.owner = owner
        self.name = name
        self._name = f'_{name}'
        self._errmsg1 = f"expected {self.type.__name__!r} but got ""{!r}"
        self._errmsg2 = f" for {self.name!r} attribute of ""{!r} object"

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        try:
            return getattr(obj, self._name)
        except AttributeError as e:
            raise type(e)(str(e).replace(self._name, self.name)) from None

    def __set__(self, obj, val):
        if not isinstance(val, self.type):
            errmsg2 = self._errmsg2.format((type(obj).__name__))
            try:
                self.type.validate(val)
            except AttributeError:
                errmsg1 = self._errmsg1.format(type(val).__name__)
                raise TypeError(errmsg1 + errmsg2)
            except (TypeError, ValueError) as err:
                raise type(err)(str(err) + errmsg2) from None
        setattr(obj, self._name, val)

    def __repr__(self):
        cls = self.__class__.__name__
        descr = self.name
        owner = self.owner.__name__
        type = self.type.__name__
        return (f"{cls}(owner={owner}, name={descr!r}, type={type})")
